/*
 * Copyright (c) 1993 by Sun Microsystems, Inc.
 */

#ifndef _SYS_KMEM_IMPL_H
#define	_SYS_KMEM_IMPL_H

#pragma ident	"@(#)kmem_impl.h	1.4	95/01/11 SMI"

#if ( defined(_CRS) && defined(sun4d) && !defined(_CRS_BIGMEM) )
#define _CRS_BIGMEM 
#endif

#include <sys/kmem.h>
#include <sys/thread.h>
#include <sys/t_lock.h>
#include <sys/time.h>
#include <sys/kstat.h>

#ifdef	__cplusplus
extern "C" {
#endif

/*
 * kernel memory allocator: implementation-private data structures
 */

extern int kmem_flags;			/* debugging features and state info */

#define	KMF_AUDIT	0x00000001	/* transaction auditing */
#define	KMF_DEADBEEF	0x00000002	/* deadbeef checking */
#define	KMF_REDZONE	0x00000004	/* redzone checking */
#define	KMF_UNMAP	0x00000008	/* synchronous slab destroy */
#define	KMF_VERIFY	0x00000010	/* freed-address verification */
#define	KMF_PAGEPERBUF	0x00000020	/* give each buf its own mapping */
#define	KMF_NOCOLOR	0x00000040	/* disable slab coloring */

#define	KMF_READY	0x00000100	/* cache is ready */
#define	KMF_HASH	0x00000200	/* cache has hash table */
#define	KMF_RESCALABLE	0x00000400	/* hash table currently rescalable */

#define	KMEM_STACK_DEPTH	15
#define	KMEM_LOG_SIZE		0x1000	/* default transaction log size */

/*
 * The bufctl (buffer control) structure keeps some minimal information
 * about each buffer: its address, its slab, and its current linkage,
 * which is either on the slab's freelist (if the buffer is free), or
 * on the cache's buf-to-bufctl hash table (if the buffer is allocated).
 * In the case of non-hashed, or "raw", caches (the common case), only
 * the freelist linkage is necessary: the buffer address is at a fixed
 * offset from the bufctl address, and the slab is at the end of the page.
 *
 * NOTE: bc_next must be the first field; raw buffers have linkage only.
 */
typedef struct kmem_bufctl {
	struct kmem_bufctl	*bc_next;	/* next bufctl struct */
	void			*bc_addr;	/* address of buffer */
	struct kmem_slab	*bc_slab;	/* controlling slab */
} kmem_bufctl_t;

/*
 * The KMF_AUDIT version of the bufctl structure.  The beginning of this
 * structure must be identical to the normal bufctl structure, so that
 * pointers are interchangeable.
 */
typedef struct kmem_bufctl_audit {
	struct kmem_bufctl	*bc_next;	/* next bufctl struct */
	void			*bc_addr;	/* address of buffer */
	struct kmem_slab	*bc_slab;	/* controlling slab */
	struct kmem_cache	*bc_cache;	/* controlling cache */
	struct kmem_bufctl_audit *bc_lastlog;	/* last log entry */
	kthread_id_t		bc_thread;	/* thread doing transaction */
	hrtime_t		bc_timestamp;	/* transaction time */
	int			bc_depth;	/* stack depth */
	u_int			bc_stack[KMEM_STACK_DEPTH];	/* pc stack */
} kmem_bufctl_audit_t;

typedef struct kmem_slab {
	struct kmem_cache	*slab_cache;	/* controlling cache */
	void			*slab_base;	/* base of allocated memory */
	struct kmem_slab	*slab_next;	/* next slab on freelist */
	struct kmem_slab	*slab_prev;	/* prev slab on freelist */
	struct kmem_bufctl	*slab_head;	/* first free buffer */
	struct kmem_bufctl	*slab_tail;	/* last free buffer */
	int			slab_refcnt;	/* outstanding allocations */
	int			slab_chunks;	/* chunks (bufs) in this slab */
} kmem_slab_t;

#define	KMEM_MIN_HASH_SIZE	64

#define	KMEM_HASH(buf, hp)	\
	((hp)->hash_table +	\
	(((u_int)(buf) >> (hp)->hash_shift) & (hp)->hash_mask))

typedef struct kmem_hash {
	int		hash_shift;	/* where the interesting bits start */
	int		hash_mask;	/* hash_mask + 1 == table size  */
	kmem_bufctl_t	**hash_table;	/* hash table base */
} kmem_hash_t;

#define	KMEM_CACHE_NAMELEN	31

typedef struct kmem_cache {
	kmutex_t	cache_lock;	/* protects this cache only */
	int		cache_flags;	/* various cache state info */
	kmem_slab_t	*cache_freelist;	/* slab free list */
	int		cache_offset;	/* for buf-to-bufctl conversion */
	int		cache_alloc;	/* total allocations */
	int		cache_alloc_fail;	/* failed allocations */
	kmem_hash_t	cache_hash;	/* hash table */
	kmem_slab_t	cache_nullslab;		/* end of freelist marker */
	kmem_slab_t	*cache_deathrow;	/* reclaimable slabs */
	int		cache_deathrow_size;	/* slabs on deathrow */
	int		cache_deathrow_reaplimit; /* working-set throttle */
	void		(*cache_constructor)(void *, size_t);
	void		(*cache_destructor)(void *, size_t);
	int		cache_bufsize;	/* size of buffers in this cache */
	int		cache_align;	/* minimum guaranteed alignment */
	int		cache_chunksize;	/* buf + alignment (+ debug) */
	int		cache_slabsize;	/* size of a slab */
	int		cache_color;	/* color to assign to next new slab */
	int		cache_maxcolor;	/* maximum slab color */
	int		cache_slab_create;	/* total slab creates */
	int		cache_slab_destroy;	/* total slab destroys */
	int		cache_buftotal;	/* total buffers (alloc + avail) */
	int		cache_bufmax;	/* max buffers ever in this cache */
	int		cache_rescale;	/* # of hash table rescales */
	int		cache_lookup_depth;	/* hash lookup depth */
	kmem_hash_t	cache_prevhash;	/* previous hash table */
	kstat_t		*cache_kstat;	/* statistics */
	kcondvar_t	cache_cv;	/* general event notification */
	struct kmem_cache	*cache_next;	/* forward cache linkage */
	struct kmem_cache	*cache_prev;	/* backward cache linkage */
	char		cache_name[KMEM_CACHE_NAMELEN + 1];
#ifdef _CRS_BIGMEM
	void            *(*cache_page_allocator)(int, int);
	void            (*cache_page_deallocator)(void *, int);
	int             cache_mem_type;
#endif _CRS_BIGMEM
} kmem_cache_t;

typedef struct kmem_async {
	struct kmem_async	*async_next;
	struct kmem_async	*async_prev;
	void			(*async_func)(kmem_cache_t *);
	kmem_cache_t		*async_cache;
} kmem_async_t;

#define	KMEM_PERM_MINFREE	32

typedef struct kmem_perm {
	struct kmem_perm	*perm_next;
	char			*perm_base;
	char			*perm_current;
	int			perm_avail;
} kmem_perm_t;

#define	KMEM_ALIGN		8	/* min guaranteed alignment */
#define	KMEM_ALIGN_SHIFT	3	/* log2(KMEM_ALIGN) */
#define	KMEM_VOID_FRACTION	8	/* never waste more than 1/8 of slab */

#ifdef	__cplusplus
}
#endif

#endif	/* _SYS_KMEM_IMPL_H */
